# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning.model.types
import unified_planning.environment
import unified_planning.walkers as walkers
from unified_planning.model.types import BOOL
from unified_planning.model import FNode, OperatorKind
from unified_planning.exceptions import UPTypeError
from typing import List, Optional


class TypeChecker(walkers.DagWalker):
    def __init__(self, env: 'unified_planning.environment.Environment'):
        walkers.DagWalker.__init__(self)
        self.env = env

    def get_type(self, expression: FNode) -> 'unified_planning.model.types.Type':
        """ Returns the unified_planning.model.types type of the expression """
        res = self.walk(expression)
        if res is None:
            raise UPTypeError("The expression '%s' is not well-formed" \
                               % str(expression))
        return res

    def is_compatible_type(self, t_left: 'unified_planning.model.types.Type', t_right: 'unified_planning.model.types.Type') -> bool:
        if t_left == t_right:
            return True
        if t_left.is_user_type() and t_right.is_user_type():
            return t_right in self.env.type_manager.user_type_ancestors(t_left)
        if not ((t_left.is_int_type() and t_right.is_int_type()) or
                (t_left.is_real_type() and t_right.is_real_type()) or
                (t_left.is_real_type() and t_right.is_int_type())):
            return False
        left_lower = -float('inf') if t_left.lower_bound is None else t_left.lower_bound # type: ignore
        left_upper = float('inf') if t_left.upper_bound is None else t_left.upper_bound # type: ignore
        right_lower = -float('inf') if t_right.lower_bound is None else t_right.lower_bound # type: ignore
        right_upper = float('inf') if t_right.upper_bound is None else t_right.upper_bound # type: ignore
        if right_upper < left_lower or right_lower > left_upper:
            return False
        else:
            return True

    def is_compatible_exp(self, fluent_exp: FNode, value_exp: FNode) -> bool:
        """Returns true iff the given expressions have compatible types."""
        t_left = self.get_type(fluent_exp)
        t_right = self.get_type(value_exp)
        return self.is_compatible_type(t_left, t_right)

    @walkers.handles(OperatorKind.AND, OperatorKind.OR, OperatorKind.NOT, OperatorKind.IMPLIES, OperatorKind.IFF, OperatorKind.EXISTS, OperatorKind.FORALL)
    def walk_bool_to_bool(self, expression: FNode,
                          args: List['unified_planning.model.types.Type']) -> Optional['unified_planning.model.types.Type']:
        assert expression is not None
        for x in args:
            if x is None or x != BOOL:
                return None
        return BOOL

    def walk_fluent_exp(self, expression: FNode, args: List['unified_planning.model.types.Type']) -> Optional['unified_planning.model.types.Type']:
        assert expression.is_fluent_exp()
        f = expression.fluent()
        if len(args) != len(f.signature):
            return None
        for (arg, param) in zip(args, f.signature):
            if not self.is_compatible_type(arg, param.type):
                return None
        return f.type

    def walk_param_exp(self, expression: FNode, args: List['unified_planning.model.types.Type']) -> 'unified_planning.model.types.Type':
        assert expression is not None
        assert len(args) == 0
        return expression.parameter().type

    def walk_variable_exp(self, expression: FNode, args: List['unified_planning.model.types.Type']) -> 'unified_planning.model.types.Type':
        assert expression is not None
        assert len(args) == 0
        return expression.variable().type

    def walk_object_exp(self, expression: FNode, args: List['unified_planning.model.types.Type']) -> 'unified_planning.model.types.Type':
        assert expression is not None
        assert len(args) == 0
        return expression.object().type

    @walkers.handles(OperatorKind.BOOL_CONSTANT)
    def walk_identity_bool(self, expression: FNode,
                           args: List['unified_planning.model.types.Type']) -> Optional['unified_planning.model.types.Type']:
        assert expression is not None
        assert len(args) == 0
        return BOOL

    @walkers.handles(OperatorKind.REAL_CONSTANT)
    def walk_identity_real(self, expression, args):
        assert expression is not None
        assert len(args) == 0
        return self.env.type_manager.RealType(expression.constant_value(), expression.constant_value())

    @walkers.handles(OperatorKind.INT_CONSTANT)
    def walk_identity_int(self, expression, args):
        assert expression is not None
        assert len(args) == 0
        return self.env.type_manager.IntType(expression.constant_value(), expression.constant_value())

    def walk_plus(self, expression, args):
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        for x in args:
            if x.lower_bound is None:
                lower = -float('inf')
            elif lower is None:
                lower = x.lower_bound
            else:
                lower += x.lower_bound
            if x.upper_bound is None:
                upper = float('inf')
            elif upper is None:
                upper = x.upper_bound
            else:
                upper += x.upper_bound
        if lower == -float('inf'):
            lower = None
        if upper == float('inf'):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_minus(self, expression, args):
        assert len(args) == 2
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        left = args[0]
        right = args[1]
        left_lower = -float('inf') if left.lower_bound is None else left.lower_bound
        left_upper = float('inf') if left.upper_bound is None else left.upper_bound
        right_lower = -float('inf') if right.lower_bound is None else right.lower_bound
        right_upper = float('inf') if right.upper_bound is None else right.upper_bound
        lower = left_lower - right_upper
        upper = left_upper - right_lower
        if lower == -float('inf'):
            lower = None
        if upper == float('inf'):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_times(self, expression, args):
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        for x in args:
            l = -float('inf') if x.lower_bound is None else x.lower_bound
            u = float('inf') if x.upper_bound is None else x.upper_bound
            if lower is None:
                lower = l
                upper = u
            else:
                lower = min(lower*l, lower*u, upper*l, upper*u)
                upper = max(lower*l, lower*u, upper*l, upper*u)
        if lower == -float('inf'):
            lower = None
        if upper == float('inf'):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_div(self, expression, args):
        assert len(args) == 2
        has_real = False
        to_skip = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
            if x.lower_bound is None and x.upper_bound is None:
                to_skip = True
        left = args[0]
        right = args[1]
        if to_skip or right.lower_bound != right.upper_bound:
            pass
        else:
            left_lower = -float('inf') if left.lower_bound is None else left.lower_bound
            left_upper = float('inf') if left.upper_bound is None else left.upper_bound
            right = right.lower_bound
            lower = min(left_lower/right, left_upper/right)
            upper = max(left_lower/right, left_upper/right)
        if lower == -float('inf'):
            lower = None
        if upper == float('inf'):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    @walkers.handles(OperatorKind.LE, OperatorKind.LT)
    def walk_math_relation(self, expression, args):
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
        return BOOL

    def walk_equals(self, expression: FNode,
                    args: List['unified_planning.model.types.Type']) -> Optional['unified_planning.model.types.Type']:
        t = args[0]
        if t is None:
            return None

        if t.is_bool_type():
            raise UPTypeError("The expression '%s' is not well-formed."
                               "Equality operator is not supported for Boolean"
                               " terms. Use Iff instead." \
                               % str(expression))
        for x in args:
            if x is None:
                return None
            elif t.is_user_type() and t != x:
                return None
            elif (t.is_int_type() or t.is_real_type()) and not (x.is_int_type() or x.is_real_type()):
                return None
        return BOOL



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import classical_kind, basic_numeric_kind, quality_metrics_kind
from unified_planning.test import TestCase, main, skipIfSolverNotAvailable
from unified_planning.test import skipIfNoOneshotPlannerForProblemKind
from unified_planning.test import skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee
from unified_planning.test.examples import get_example_problems
from unified_planning.solvers import PlanGenerationResultStatus
from unified_planning.solvers.results import POSITIVE_OUTCOMES


class TestPlanner(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfSolverNotAvailable('tamer')
    def test_basic(self):
        problem = self.problems['basic'].problem
        a = problem.action('a')

        with OneshotPlanner(name='tamer', params={'weight': 0.8}) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfSolverNotAvailable('tamer')
    def test_basic_with_timeout(self):
        problem = self.problems['basic'].problem
        a = problem.action('a')

        with OneshotPlanner(name='tamer', params={'weight': 0.8}) as planner:
            self.assertNotEqual(planner, None)
            with warnings.catch_warnings(record=True) as w:
                final_report = planner.solve(problem, timeout = 0.001)
                self.assertIn(final_report.status, POSITIVE_OUTCOMES)
                plan = final_report.plan
                self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING)
                self.assertEqual(len(plan.actions), 1)
                self.assertEqual(plan.actions[0].action, a)
                self.assertEqual(len(plan.actions[0].actual_parameters), 0)
                self.assertEqual(len(w), 1)
                self.assertEqual('Tamer does not support timeout.', str(w[-1].message))

    @skipIfSolverNotAvailable('tamer')
    def test_basic_parallel(self):
        problem = self.problems['basic'].problem
        a = problem.action('a')

        with OneshotPlanner(names=['tamer', 'tamer'],
                            params=[{'heuristic': 'hadd'}, {'heuristic': 'hmax'}]) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(quality_metrics_kind))
    @skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee(PlanGenerationResultStatus.SOLVED_OPTIMALLY)
    def test_actions_cost(self):
        problem = self.problems['basic_with_costs'].problem
        opt_plan = self.problems['basic_with_costs'].plan
        with OneshotPlanner(problem_kind=problem.kind,
                            optimality_guarantee=PlanGenerationResultStatus.SOLVED_OPTIMALLY) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY)
            self.assertEqual(plan, opt_plan)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_numeric_kind))
    def test_robot(self):
        problem = self.problems['robot'].problem
        move = problem.action('move')

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    def test_robot_loader(self):
        problem = self.problems['robot_loader'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(len(plan.actions), 4)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)
            self.assertEqual(len(plan.actions[1].actual_parameters), 1)
            self.assertEqual(len(plan.actions[2].actual_parameters), 2)
            self.assertEqual(len(plan.actions[3].actual_parameters), 1)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    def test_robot_loader_adv(self):
        problem = self.problems['robot_loader_adv'].problem
        move = problem.action('move')
        load = problem.action('load')
        unload = problem.action('unload')

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)

if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test.examples import get_example_problems
from unified_planning.test import TestCase, main


class TestModel(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_clone_problem_and_action(self):
        for _, (problem, _) in self.problems.items():
            problem_clone_1 = problem.clone()
            problem_clone_2 = problem.clone()
            for action_1, action_2 in zip(problem_clone_1.actions, problem_clone_2.actions):
                if isinstance(action_2, InstantaneousAction):
                    action_2._effects = []
                    action_1_clone = action_1.clone()
                    action_1_clone._effects = []
                elif isinstance(action_2, DurativeAction):
                    action_2._effects = {}
                    action_1_clone = action_1.clone()
                    action_1_clone._effects = {}
                else:
                    raise NotImplementedError
                self.assertEqual(action_2, action_1_clone)
                self.assertEqual(action_1_clone, action_2)
                self.assertNotEqual(action_1, action_1_clone)
                self.assertNotEqual(action_1_clone, action_1)
                self.assertNotEqual(action_1, action_1_clone.name)
                self.assertNotEqual(action_1_clone.name, action_1)
            self.assertEqual(problem_clone_1, problem)
            self.assertEqual(problem, problem_clone_1)
            self.assertNotEqual(problem_clone_2, problem)
            self.assertNotEqual(problem, problem_clone_2)

    def test_clone_action(self):
        Location = UserType('Location')
        a = Action('move', l_from=Location, l_to=Location)
        with self.assertRaises(NotImplementedError):
            a.clone()
        with self.assertRaises(NotImplementedError):
            hash(a)
        with self.assertRaises(NotImplementedError):
            a == a.name
        with self.assertRaises(NotImplementedError):
            a.is_conditional()

    def test_clone_effect(self):
        x = FluentExp(Fluent('x'))
        y = FluentExp(Fluent('y'))
        z = FluentExp(Fluent('z'))
        e = Effect(x, z, y, unified_planning.model.EffectKind.ASSIGN)
        e_clone_1 = e.clone()
        e_clone_2 = e.clone()
        e_clone_2._condition = TRUE()
        self.assertEqual(e_clone_1, e)
        self.assertEqual(e, e_clone_1)
        self.assertNotEqual(e_clone_2, e)
        self.assertNotEqual(e, e_clone_2)
        self.assertNotEqual(e, e.value)
        self.assertNotEqual(e.value, e)

    def test_istantaneous_action(self):
        Location = UserType('Location')
        move = InstantaneousAction('move', l_from=Location, l_to=Location)
        km = Fluent('km', IntType())
        move.add_increase_effect(km, 10)
        e = Effect(FluentExp(km), Int(10), TRUE(), unified_planning.model.EffectKind.INCREASE)
        self.assertEqual(move.effects[0], e)

    def test_durative_action(self):
        Location = UserType('Location')
        x = Fluent('x')
        move = DurativeAction('move', l_from=Location, l_to=Location)
        km = Fluent('km', IntType())
        move.add_decrease_effect(StartTiming(), km, 5)
        move.add_increase_effect(EndTiming(), km, 20)
        e_end = Effect(FluentExp(km), Int(20), TRUE(), unified_planning.model.EffectKind.INCREASE)
        e_start = Effect(FluentExp(km), Int(5), TRUE(), unified_planning.model.EffectKind.DECREASE)
        effects_test = {StartTiming(): [e_start], EndTiming(): [e_end]}
        self.assertEqual(effects_test, move.effects)
        move.set_closed_duration_interval(1, 2)
        self.assertEqual(move.duration, ClosedDurationInterval(Int(1), Int(2)))
        move.set_open_duration_interval(2, Fraction(7, 2))
        self.assertEqual(move.duration, OpenDurationInterval(Int(2), Real(Fraction(7, 2))))
        move.set_left_open_duration_interval(1, 2)
        self.assertEqual(move.duration, LeftOpenDurationInterval(Int(1), Int(2)))
        move.set_right_open_duration_interval(1, 2)
        self.assertEqual(move.duration, RightOpenDurationInterval(Int(1), Int(2)))
        move.add_condition(StartTiming(), x)
        move.add_condition(ClosedTimeInterval(StartTiming(), EndTiming()), x)
        self.assertIn('duration = [1, 2)', str(move))

    def test_problem(self):
        x = Fluent('x')
        y = Fluent('y')
        km = Fluent('km', IntType())
        problem = Problem('problem_test')
        problem.add_fluent(x, default_initial_value=True)
        problem.add_fluent(y, default_initial_value=False)
        problem.add_fluent(km, default_initial_value=Int(0))
        problem.add_timed_effect(GlobalStartTiming(5), x, y)
        problem.add_timed_goal(GlobalStartTiming(11), x)
        problem.add_timed_goal(TimeInterval(GlobalStartTiming(5), GlobalStartTiming(9)), x)
        problem.add_action(DurativeAction('move'))
        problem.add_action(InstantaneousAction('stop_moving'))
        stop_moving_list = [a for a in problem.instantaneous_actions]
        self.assertEqual(len(stop_moving_list), 1)
        stop_moving = stop_moving_list[0]
        self.assertEqual(stop_moving.name, 'stop_moving')
        move_list = [a for a in problem.durative_actions]
        self.assertEqual(len(move_list), 1)
        move = move_list[0]
        self.assertEqual(move.name, 'move')
        problem.add_increase_effect(GlobalStartTiming(5), km, 10)
        problem.add_decrease_effect(GlobalStartTiming(10), km, 5)
        self.assertIn(str(Effect(FluentExp(km), Int(10), TRUE(), unified_planning.model.EffectKind.INCREASE)), str(problem))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module implements the grounder that uses tarski."""


from functools import partial
import shutil
from typing import Callable, Optional, Tuple, Dict, List
import tarski # type: ignore
import unified_planning
import unified_planning.interop
from unified_planning.interop.from_tarski import convert_tarski_formula
from unified_planning.model.problem_kind import full_classical_kind, full_numeric_kind
from unified_planning.model import Action, FNode
from unified_planning.solvers.grounder import lift_plan
from unified_planning.solvers.solver import Solver
from unified_planning.transformers import Grounder
from tarski.grounding import LPGroundingStrategy # type: ignore


gringo = shutil.which('gringo')
if gringo is None:
    raise ImportError('Tarski grounder needs gringo installed')


class TarskiGrounder(Solver):
    """Implements the gounder that uses tarski."""
    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise

    @property
    def name(self) -> str:
        return 'tarski_grounder'

    @staticmethod
    def is_grounder() -> bool:
        return True

    @staticmethod
    def supports(problem_kind: 'unified_planning.model.ProblemKind') -> bool:
        supported_kind = unified_planning.model.ProblemKind()
        supported_kind.set_typing('FLAT_TYPING') # type: ignore
        supported_kind.set_typing('HIERARCHICAL_TYPING') # type: ignore
        supported_kind.set_conditions_kind('NEGATIVE_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('DISJUNCTIVE_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('EQUALITY') # type: ignore
        supported_kind.set_conditions_kind('EXISTENTIAL_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('UNIVERSAL_CONDITIONS') # type: ignore
        supported_kind.set_effects_kind('CONDITIONAL_EFFECTS') # type: ignore
        return problem_kind <= supported_kind

    def ground(self, problem: 'unified_planning.model.Problem') -> Tuple['unified_planning.model.Problem', Callable[[unified_planning.plan.Plan], unified_planning.plan.Plan]]:
        tarski_problem = unified_planning.interop.convert_problem_to_tarski(problem)
        actions = None
        try:
            lpgs = LPGroundingStrategy(tarski_problem)
            actions = lpgs.ground_actions()
        except tarski.grounding.errors.ReachabilityLPUnsolvable:
            raise unified_planning.exceptions.UPUsageError('tarski grounder can not find a solvable grounding.')
        grounded_actions_map: Dict[Action, List[Tuple[FNode, ...]]] = {}
        fluents = {fluent.name: fluent for fluent in problem.fluents}
        objects = {object.name: object for object in problem.all_objects}
        types: Dict[str, Optional['unified_planning.model.Type']] = {}
        if not problem.has_type('object'):
            types['object'] = None # we set object as None, so when it is the father of a type in tarski, in UP it will be None.
        for action_name, list_of_tuple_of_parameters in actions.items():
            action = problem.action(action_name)
            parameters = {parameter.name: parameter for parameter in action.parameters}
            grounded_actions_map[action] = []
            for tuple_of_parameters in list_of_tuple_of_parameters:
                temp_list_of_converted_parameters = []
                for p in tuple_of_parameters:
                    if isinstance(p, str):
                        temp_list_of_converted_parameters.append(problem.env.expression_manager.ObjectExp(problem.object(p)))
                    else:
                        temp_list_of_converted_parameters.append(convert_tarski_formula(problem.env, fluents, \
                            objects, parameters, types, p))
                grounded_actions_map[action].append(tuple(temp_list_of_converted_parameters))
        unified_planning_grounder = Grounder(problem, grounding_actions_map=grounded_actions_map)
        grounded_problem = unified_planning_grounder.get_rewritten_problem()
        trace_back_map = unified_planning_grounder.get_rewrite_back_map()
        return (grounded_problem, partial(lift_plan, map=trace_back_map))

    def destroy(self):
        pass



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unified_planning as up
from unified_planning.model.problem_kind import ProblemKind
from unified_planning.environment import get_env
from typing import List, Optional, Union
from unified_planning.solvers import PDDLSolver


FILE_PATH = os.path.dirname(os.path.abspath(__file__))


class ENHSP(PDDLSolver):
    def __init__(self):
        PDDLSolver.__init__(self, False)

    @property
    def name(self) -> str:
        return 'ENHSP'

    def _get_cmd(self, domanin_filename: str, problem_filename: str, plan_filename: str) -> List[str]:
        return ['java', '-jar',
                os.path.join(FILE_PATH, '..', '..', '..', '.planners', 'enhsp-20', 'enhsp.jar'),
                '-o', domanin_filename, '-f', problem_filename, '-sp', plan_filename,
                '-planner', 'opt-hrmax']

    def _result_status(self, problem: 'up.model.Problem', plan: Optional['up.plan.Plan']) -> 'up.solvers.results.PlanGenerationResultStatus':
        if plan is None:
            return up.solvers.results.PlanGenerationResultStatus.UNSOLVABLE_PROVEN
        else:
            return up.solvers.results.PlanGenerationResultStatus.SOLVED_OPTIMALLY

    @staticmethod
    def satisfies(optimality_guarantee: Union[up.solvers.solver.OptimalityGuarantee, str]) -> bool:
        return True

    @staticmethod
    def supports(problem_kind: 'ProblemKind') -> bool:
        supported_kind = ProblemKind()
        supported_kind.set_numbers('DISCRETE_NUMBERS') # type: ignore
        supported_kind.set_numbers('CONTINUOUS_NUMBERS') # type: ignore
        supported_kind.set_typing('FLAT_TYPING') # type: ignore
        supported_kind.set_conditions_kind('NEGATIVE_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('DISJUNCTIVE_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('EQUALITY') # type: ignore
        supported_kind.set_conditions_kind('EXISTENTIAL_CONDITIONS') # type: ignore
        supported_kind.set_conditions_kind('UNIVERSAL_CONDITIONS') # type: ignore
        supported_kind.set_effects_kind('CONDITIONAL_EFFECTS') # type: ignore
        supported_kind.set_effects_kind('INCREASE_EFFECTS') # type: ignore
        supported_kind.set_effects_kind('DECREASE_EFFECTS') # type: ignore
        supported_kind.set_fluents_type('NUMERIC_FLUENTS') # type: ignore
        supported_kind.set_quality_metrics('ACTIONS_COST') # type: ignore
        supported_kind.set_quality_metrics('FINAL_VALUE') # type: ignore
        return problem_kind <= supported_kind


env = get_env()
if os.path.isfile(os.path.join(FILE_PATH, '..', '..', '..', '.planners', 'enhsp-20', 'enhsp.jar')):
    env.factory.add_solver('opt-pddl-solver', 'unified_planning.test.pddl.enhsp', 'ENHSP')



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning.test.examples.minimals as minimals
import unified_planning.test.examples.realistic as realistic
import unified_planning.test.examples.testing_variants as testing_variants


def get_example_problems():
    sub_modules = [minimals, realistic, testing_variants]
    return dict(x for m in sub_modules for x in m.get_example_problems().items())



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.test.examples import get_example_problems
from typing import OrderedDict


class TestProblem(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_problem_kind(self):
        problem_kind = ProblemKind()
        self.assertFalse(problem_kind.has_discrete_time())
        self.assertFalse(problem_kind.has_continuous_time())
        problem_kind.set_time('DISCRETE_TIME')
        self.assertTrue(problem_kind.has_discrete_time())
        problem_kind.set_time('CONTINUOUS_TIME')
        self.assertTrue(problem_kind.has_continuous_time())

    def test_basic(self):
        problem = self.problems['basic'].problem

        x = problem.fluent('x')
        self.assertEqual(x.name, 'x')
        self.assertEqual(str(x), 'bool x')
        self.assertEqual(x.arity, 0)
        self.assertTrue(x.type.is_bool_type())

        a = problem.action('a')
        self.assertEqual(a.name, 'a')
        self.assertEqual(len(a.preconditions), 1)
        self.assertEqual(len(a.effects), 1)
        a_str = str(a)
        self.assertIn('action a', a_str)
        self.assertIn('preconditions', a_str)
        self.assertIn('not x', a_str)
        self.assertIn('effects', a_str)
        self.assertIn('x := true', a_str)

        self.assertEqual(problem.name, 'basic')
        self.assertEqual(len(problem.fluents), 1)
        self.assertEqual(len(problem.actions), 1)
        self.assertTrue(problem.initial_value(x) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertIn('fluents', problem_str)
        self.assertIn('actions', problem_str)
        self.assertIn('initial values', problem_str)
        self.assertIn('goals', problem_str)

    def test_basic_conditional(self):
        problem = self.problems['basic_conditional'].problem

        x = problem.fluent('x')
        self.assertEqual(x.name, 'x')
        self.assertEqual(str(x), 'bool x')
        self.assertEqual(x.arity, 0)
        self.assertTrue(x.type.is_bool_type())

        y = problem.fluent('y')
        self.assertEqual(y.name, 'y')
        self.assertEqual(str(y), 'bool y')
        self.assertEqual(y.arity, 0)
        self.assertTrue(y.type.is_bool_type())

        a_x = problem.action('a_x')
        self.assertEqual(a_x.name, 'a_x')
        self.assertEqual(len(a_x.preconditions), 1)
        self.assertEqual(len(a_x.effects), 1)
        ax_str = str(a_x)
        self.assertIn('action a_x', ax_str)
        self.assertIn('preconditions', ax_str)
        self.assertIn('not x', ax_str)
        self.assertIn('effects', ax_str)
        self.assertIn('if y then x := true', ax_str)

        a_y = problem.action('a_y')
        self.assertEqual(a_y.name, 'a_y')
        self.assertEqual(len(a_y.preconditions), 1)
        self.assertEqual(len(a_y.effects), 1)
        ay_str = str(a_y)
        self.assertIn('action a_y', ay_str)
        self.assertIn('preconditions', ay_str)
        self.assertIn('not y', ay_str)
        self.assertIn('effects', ay_str)
        self.assertIn('y := true', ay_str)

        self.assertEqual(problem.name, 'basic_conditional')
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(len(problem.actions), 2)
        self.assertTrue(problem.initial_value(x) is not None)
        self.assertTrue(problem.initial_value(y) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertIn('fluents', problem_str)
        self.assertIn('actions', problem_str)
        self.assertIn('initial values', problem_str)
        self.assertIn('goals', problem_str)

    def test_robot(self):
        problem = self.problems['robot'].problem

        Location = problem.user_type('Location')
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, 'Location')
        self.assertEqual(str(Location), 'Location')

        robot_at = problem.fluent('robot_at')
        self.assertEqual(robot_at.name, 'robot_at')
        self.assertEqual(str(robot_at), 'bool robot_at[position=Location]')
        self.assertEqual(robot_at.arity, 1)
        self.assertEqual(robot_at.signature, [up.model.Parameter('position', Location)])
        self.assertTrue(robot_at.type.is_bool_type())

        battery_charge = problem.fluent('battery_charge')
        self.assertEqual(battery_charge.name, 'battery_charge')
        self.assertEqual(str(battery_charge), 'real[0, 100] battery_charge')
        self.assertEqual(battery_charge.arity, 0)
        self.assertTrue(battery_charge.type.is_real_type())

        move = problem.action('move')
        l_from = move.parameter('l_from')
        l_to = move.parameter('l_to')
        self.assertEqual(move.name, 'move')
        self.assertEqual(len(move.parameters), 2)
        self.assertEqual(l_from.name, 'l_from')
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, 'l_to')
        self.assertEqual(l_to.type, Location)
        self.assertEqual(len(move.preconditions), 4)
        self.assertEqual(len(move.effects), 3)
        move_str = str(move)
        self.assertTrue('action move(Location l_from, Location l_to)' in move_str)
        self.assertTrue('preconditions' in move_str)
        self.assertTrue('10 <= battery_charge' in move_str)
        self.assertTrue('not (l_from == l_to)' in move_str)
        self.assertTrue('robot_at(l_from)' in move_str)
        self.assertTrue('not robot_at(l_to)' in move_str)
        self.assertTrue('effects' in move_str)
        self.assertTrue('robot_at(l_from) := false' in move_str)
        self.assertTrue('robot_at(l_to) := true' in move_str)
        self.assertTrue('battery_charge := (battery_charge - 10)' in move_str)

        l1 = problem.object('l1')
        l2 = problem.object('l2')
        self.assertEqual(l1.name, 'l1')
        self.assertEqual(str(l1), 'l1')
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, 'l2')
        self.assertEqual(str(l2), 'l2')
        self.assertEqual(l2.type, Location)

        self.assertEqual(problem.name, 'robot')
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(problem.fluent('robot_at'), robot_at)
        self.assertEqual(problem.fluent('battery_charge'), battery_charge)
        self.assertEqual(len(problem.user_types), 1)
        self.assertEqual(problem.user_type('Location'), Location)
        self.assertEqual(len(list(problem.objects(Location))), 2)
        self.assertEqual(list(problem.objects(Location)), [l1, l2])
        self.assertEqual(len(problem.actions), 1)
        self.assertEqual(problem.action('move'), move)
        self.assertTrue(problem.initial_value(robot_at(l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(l2)) is not None)
        self.assertTrue(problem.initial_value(battery_charge) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertTrue('types' in problem_str)
        self.assertTrue('fluents' in problem_str)
        self.assertTrue('actions' in problem_str)
        self.assertTrue('objects' in problem_str)
        self.assertTrue('initial values' in problem_str)
        self.assertTrue('goals' in problem_str)

    def test_robot_loader(self):
        problem = self.problems['robot_loader'].problem

        Location = problem.user_type('Location')
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, 'Location')

        robot_at = problem.fluent('robot_at')
        self.assertEqual(robot_at.name, 'robot_at')
        self.assertEqual(robot_at.arity, 1)
        self.assertEqual(robot_at.signature, [up.model.Parameter('position', Location)])
        self.assertTrue(robot_at.type.is_bool_type())

        cargo_at = problem.fluent('cargo_at')
        self.assertEqual(cargo_at.name, 'cargo_at')
        self.assertEqual(cargo_at.arity, 1)
        self.assertEqual(cargo_at.signature, [up.model.Parameter('position', Location)])
        self.assertTrue(cargo_at.type.is_bool_type())

        cargo_mounted = problem.fluent('cargo_mounted')
        self.assertEqual(cargo_mounted.name, 'cargo_mounted')
        self.assertEqual(cargo_mounted.arity, 0)
        self.assertTrue(cargo_mounted.type.is_bool_type())

        move = problem.action('move')
        l_from = move.parameter('l_from')
        l_to = move.parameter('l_to')
        self.assertEqual(move.name, 'move')
        self.assertEqual(len(move.parameters), 2)
        self.assertEqual(l_from.name, 'l_from')
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, 'l_to')
        self.assertEqual(l_to.type, Location)
        self.assertEqual(len(move.preconditions), 3)
        self.assertEqual(len(move.effects), 2)

        load = problem.action('load')
        loc = load.parameter('loc')
        self.assertEqual(load.name, 'load')
        self.assertEqual(len(load.parameters), 1)
        self.assertEqual(loc.name, 'loc')
        self.assertEqual(loc.type, Location)
        self.assertEqual(len(load.preconditions), 3)
        self.assertEqual(len(load.effects), 2)

        unload = problem.action('unload')
        loc = unload.parameter('loc')
        self.assertEqual(unload.name, 'unload')
        self.assertEqual(len(unload.parameters), 1)
        self.assertEqual(loc.name, 'loc')
        self.assertEqual(loc.type, Location)
        self.assertEqual(len(unload.preconditions), 3)
        self.assertEqual(len(unload.effects), 2)

        l1 = problem.object('l1')
        l2 = problem.object('l2')
        self.assertEqual(l1.name, 'l1')
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, 'l2')
        self.assertEqual(l2.type, Location)

        self.assertEqual(problem.name, 'robot_loader')
        self.assertEqual(len(problem.fluents), 3)
        self.assertEqual(problem.fluent('robot_at'), robot_at)
        self.assertEqual(problem.fluent('cargo_at'), cargo_at)
        self.assertEqual(problem.fluent('cargo_mounted'), cargo_mounted)
        self.assertEqual(len(problem.user_types), 1)
        self.assertEqual(problem.user_type('Location'), Location)
        self.assertEqual(len(list(problem.objects(Location))), 2)
        self.assertEqual(list(problem.objects(Location)), [l1, l2])
        self.assertEqual(len(problem.actions), 3)
        self.assertEqual(problem.action('move'), move)
        self.assertEqual(problem.action('load'), load)
        self.assertEqual(problem.action('unload'), unload)
        self.assertTrue(problem.initial_value(robot_at(l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(l1)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_mounted) is not None)
        self.assertEqual(len(problem.goals), 1)

    def test_robot_loader_adv(self):
        problem = self.problems['robot_loader_adv'].problem

        Location = problem.user_type('Location')
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, 'Location')

        Robot = problem.user_type('Robot')
        self.assertTrue(Robot.is_user_type())
        self.assertEqual(Robot.name, 'Robot')

        Container = problem.user_type('Container')
        self.assertTrue(Container.is_user_type())
        self.assertEqual(Container.name, 'Container')

        robot_at = problem.fluent('robot_at')
        self.assertEqual(robot_at.name, 'robot_at')
        self.assertEqual(robot_at.arity, 2)
        self.assertEqual(robot_at.signature, [up.model.Parameter('robot', Robot), up.model.Parameter('position', Location)])
        self.assertTrue(robot_at.type.is_bool_type())

        cargo_at = problem.fluent('cargo_at')
        self.assertEqual(cargo_at.name, 'cargo_at')
        self.assertEqual(cargo_at.arity, 2)
        self.assertEqual(cargo_at.signature, [up.model.Parameter('cargo', Container), up.model.Parameter('position', Location)])
        self.assertTrue(cargo_at.type.is_bool_type())

        cargo_mounted = problem.fluent('cargo_mounted')
        self.assertEqual(cargo_mounted.name, 'cargo_mounted')
        self.assertEqual(cargo_mounted.arity, 2)
        self.assertEqual(cargo_mounted.signature, [up.model.Parameter('cargo', Container), up.model.Parameter('robot', Robot)])
        self.assertTrue(cargo_mounted.type.is_bool_type())

        move = problem.action('move')
        l_from = move.parameter('l_from')
        l_to = move.parameter('l_to')
        r = move.parameter('r')
        self.assertEqual(move.name, 'move')
        self.assertEqual(len(move.parameters), 3)
        self.assertEqual(l_from.name, 'l_from')
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, 'l_to')
        self.assertEqual(l_to.type, Location)
        self.assertEqual(r.name, 'r')
        self.assertEqual(r.type, Robot)
        self.assertEqual(len(move.preconditions), 3)
        self.assertEqual(len(move.effects), 2)

        load = problem.action('load')
        loc = load.parameter('loc')
        r = load.parameter('r')
        c = load.parameter('c')
        self.assertEqual(load.name, 'load')
        self.assertEqual(len(load.parameters), 3)
        self.assertEqual(loc.name, 'loc')
        self.assertEqual(loc.type, Location)
        self.assertEqual(r.name, 'r')
        self.assertEqual(r.type, Robot)
        self.assertEqual(c.name, 'c')
        self.assertEqual(c.type, Container)
        self.assertEqual(len(load.preconditions), 3)
        self.assertEqual(len(load.effects), 2)

        unload = problem.action('unload')
        loc = unload.parameter('loc')
        r = unload.parameter('r')
        c = unload.parameter('c')
        self.assertEqual(unload.name, 'unload')
        self.assertEqual(len(unload.parameters), 3)
        self.assertEqual(loc.name, 'loc')
        self.assertEqual(loc.type, Location)
        self.assertEqual(r.name, 'r')
        self.assertEqual(r.type, Robot)
        self.assertEqual(c.name, 'c')
        self.assertEqual(c.type, Container)
        self.assertEqual(len(unload.preconditions), 3)
        self.assertEqual(len(unload.effects), 2)

        l1 = problem.object('l1')
        l2 = problem.object('l2')
        l3 = problem.object('l3')
        r1 = problem.object('r1')
        c1 = problem.object('c1')
        self.assertEqual(l1.name, 'l1')
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, 'l2')
        self.assertEqual(l2.type, Location)
        self.assertEqual(l3.name, 'l3')
        self.assertEqual(l3.type, Location)
        self.assertEqual(r1.name, 'r1')
        self.assertEqual(r1.type, Robot)
        self.assertEqual(c1.name, 'c1')
        self.assertEqual(c1.type, Container)

        self.assertEqual(problem.name, 'robot_loader_adv')
        self.assertEqual(len(problem.fluents), 3)
        self.assertEqual(problem.fluent('robot_at'), robot_at)
        self.assertEqual(problem.fluent('cargo_at'), cargo_at)
        self.assertEqual(problem.fluent('cargo_mounted'), cargo_mounted)
        self.assertEqual(len(problem.user_types), 3)
        self.assertEqual(problem.user_type('Location'), Location)
        self.assertEqual(len(list(problem.objects(Location))), 3)
        self.assertEqual(list(problem.objects(Location)), [l1, l2, l3])
        self.assertEqual(problem.user_type('Robot'), Robot)
        self.assertEqual(len(list(problem.objects(Robot))), 1)
        self.assertEqual(list(problem.objects(Robot)), [r1])
        self.assertEqual(problem.user_type('Container'), Container)
        self.assertEqual(len(list(problem.objects(Container))), 1)
        self.assertEqual(list(problem.objects(Container)), [c1])
        self.assertEqual(len(problem.actions), 3)
        self.assertEqual(problem.action('move'), move)
        self.assertEqual(problem.action('load'), load)
        self.assertEqual(problem.action('unload'), unload)
        self.assertTrue(problem.initial_value(robot_at(r1, l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(r1, l2)) is not None)
        self.assertTrue(problem.initial_value(robot_at(r1, l3)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l1)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l3)) is not None)
        self.assertTrue(problem.initial_value(cargo_mounted(c1, r1)) is not None)
        self.assertEqual(len(problem.goals), 2)

    def test_fluents_defaults(self):
        Location = UserType('Location')
        robot_at = Fluent('robot_at', BoolType(), position=Location)
        distance = Fluent('distance', RealType(), location_1=Location, location_2=Location)

        N = 10
        locations = [Object(f'l{i}', Location) for i in range(N)]

        problem = Problem('robot')
        problem.add_fluent(robot_at, default_initial_value=False)
        problem.add_fluent(distance, default_initial_value=Fraction(-1))
        problem.add_objects(locations)
        problem.set_initial_value(robot_at(locations[0]), True)
        for i in range(N-1):
            problem.set_initial_value(distance(locations[i], locations[i+1]), Fraction(10))

        self.assertEqual(problem.initial_value(robot_at(locations[0])), TRUE())
        for i in range(1, N):
            self.assertEqual(problem.initial_value(robot_at(locations[i])), FALSE())

        for i in range(N):
            for j in range(N):
                if j == i+1:
                    self.assertEqual(problem.initial_value(distance(locations[i], locations[j])),
                                     Real(Fraction(10)))
                else:
                    self.assertEqual(problem.initial_value(distance(locations[i], locations[j])),
                                     Real(Fraction(-1)))

    def test_problem_defaults(self):
        Location = UserType('Location')
        robot_at = Fluent('robot_at', BoolType(), position=Location)
        distance = Fluent('distance', IntType(), location_1=Location, location_2=Location)
        cost = Fluent('cost', IntType(), location_1=Location, location_2=Location)

        N = 10
        locations = [Object(f'l{i}', Location) for i in range(N)]

        problem = Problem('robot', initial_defaults={IntType(): 0})
        problem.add_fluent(robot_at, default_initial_value=False)
        problem.add_fluent(distance, default_initial_value=-1)
        problem.add_fluent(cost)
        problem.add_objects(locations)
        problem.set_initial_value(robot_at(locations[0]), True)
        for i in range(N-1):
            problem.set_initial_value(distance(locations[i], locations[i+1]), 10)
            problem.set_initial_value(cost(locations[i], locations[i+1]), 100)

        self.assertEqual(problem.initial_value(robot_at(locations[0])), TRUE())
        for i in range(1, N):
            self.assertEqual(problem.initial_value(robot_at(locations[i])), FALSE())

        for i in range(N):
            for j in range(N):
                if j == i+1:
                    self.assertEqual(problem.initial_value(distance(locations[i], locations[j])), Int(10))
                    self.assertEqual(problem.initial_value(cost(locations[i], locations[j])), Int(100))
                else:
                    self.assertEqual(problem.initial_value(distance(locations[i], locations[j])), Int(-1))
                    self.assertEqual(problem.initial_value(cost(locations[i], locations[j])), Int(0))


if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import importlib
import unified_planning as up
from unified_planning.model import ProblemKind
from typing import Dict, Tuple, Optional, List, Union, Type


DEFAULT_SOLVERS = {'enhsp' : ('up_enhsp', 'ENHSPsolver'),
                   'lpg' : ('up_lpg', 'LPGsolver'),
                   'tamer' : ('up_tamer', 'SolverImpl'),
                   'pyperplan' : ('up_pyperplan', 'SolverImpl'),
                   'sequential_plan_validator' : ('unified_planning.solvers.plan_validator', 'SequentialPlanValidator'),
                   'up_grounder' : ('unified_planning.solvers.grounder', 'Grounder'),
                   'tarski_grounder' : ('unified_planning.solvers.tarski_grounder', 'TarskiGrounder')}


class Factory:
    def __init__(self, solvers: Dict[str, Tuple[str, str]] = DEFAULT_SOLVERS):
        self.solvers: Dict[str, Type['up.solvers.solver.Solver']] = {}
        for name, (module_name, class_name) in solvers.items():
            try:
                self.add_solver(name, module_name, class_name)
            except ImportError:
                pass

    def add_solver(self, name: str, module_name: str, class_name: str):
        module = importlib.import_module(module_name)
        SolverImpl = getattr(module, class_name)
        self.solvers[name] = SolverImpl

    def _get_solver_class(self, solver_kind: str, name: Optional[str] = None,
                          problem_kind: ProblemKind = ProblemKind(),
                          optimality_guarantee: Optional[Union['up.solvers.solver.OptimalityGuarantee', str]] = None) -> Optional[Type['up.solvers.solver.Solver']]:
        if name is not None:
            return self.solvers[name]
        for SolverClass in self.solvers.values():
            if getattr(SolverClass, 'is_'+solver_kind)() and SolverClass.supports(problem_kind) \
               and (optimality_guarantee is None or SolverClass.satisfies(optimality_guarantee)):
                return SolverClass
        return None

    def _get_solver(self, solver_kind: str, name: Optional[str] = None,
                    names: Optional[List[str]] = None,
                    params: Union[Dict[str, str], List[Dict[str, str]]] = None,
                    problem_kind: ProblemKind = ProblemKind(),
                    optimality_guarantee: Optional[Union['up.solvers.solver.OptimalityGuarantee', str]] = None) -> Optional['up.solvers.solver.Solver']:
        if names is not None:
            assert name is None
            if params is None:
                params = [{} for i in range(len(names))]
            assert isinstance(params, List) and len(names) == len(params)
            solvers = []
            for name, param in zip(names, params):
                SolverClass = self._get_solver_class(solver_kind, name)
                if SolverClass is None:
                    raise
                solvers.append((SolverClass, param))
            return up.solvers.parallel.Parallel(solvers)
        else:
            if params is None:
                params = {}
            assert isinstance(params, Dict)
            SolverClass = self._get_solver_class(solver_kind, name, problem_kind, optimality_guarantee)
            if SolverClass is None:
                raise
            return SolverClass(**params)
        return None

    def OneshotPlanner(self, *, name: Optional[str] = None,
                       names: Optional[List[str]] = None,
                       params: Union[Dict[str, str], List[Dict[str, str]]] = None,
                       problem_kind: ProblemKind = ProblemKind(),
                       optimality_guarantee: Optional[Union['up.solvers.solver.OptimalityGuarantee', str]] = None) -> Optional['up.solvers.solver.Solver']:
        """
        Returns a oneshot planner. There are three ways to call this method:
        - using 'name' (the name of a specific planner) and 'params' (planner dependent options).
          e.g. OneshotPlanner(name='tamer', params={'heuristic': 'hadd'})
        - using 'names' (list of specific planners name) and 'params' (list of
          planners dependent options) to get a Parallel solver.
          e.g. OneshotPlanner(names=['tamer', 'tamer'],
                              params=[{'heuristic': 'hadd'}, {'heuristic': 'hmax'}])
        - using 'problem_kind' and 'optimality_guarantee'.
          e.g. OneshotPlanner(problem_kind=problem.kind, optimality_guarantee=SOLVED_OPTIMALLY)
        """
        return self._get_solver('oneshot_planner', name, names, params, problem_kind, optimality_guarantee)

    def PlanValidator(self, *, name: Optional[str] = None,
                       names: Optional[List[str]] = None,
                       params: Union[Dict[str, str], List[Dict[str, str]]] = None,
                       problem_kind: ProblemKind = ProblemKind()) -> Optional['up.solvers.solver.Solver']:
        """
        Returns a plan validator. There are three ways to call this method:
        - using 'name' (the name of a specific plan validator) and 'params'
          (plan validator dependent options).
          e.g. PlanValidator(name='tamer', params={'opt': 'val'})
        - using 'names' (list of specific plan validators name) and 'params' (list of
          plan validators dependent options) to get a Parallel solver.
          e.g. PlanValidator(names=['tamer', 'tamer'],
                             params=[{'opt1': 'val1'}, {'opt2': 'val2'}])
        - using 'problem_kind' parameter.
          e.g. PlanValidator(problem_kind=problem.kind)
        """
        return self._get_solver('plan_validator', name, names, params, problem_kind)

    def Grounder(self, *, name: Optional[str] = None, params: Union[Dict[str, str], List[Dict[str, str]]] = None,
                       problem_kind: ProblemKind = ProblemKind()) -> Optional['up.solvers.solver.Solver']:
        """
        Returns a Grounder. There are three ways to call this method:
        - using 'name' (the name of a specific grounder) and 'params'
          (grounder dependent options).
          e.g. Grounder(name='tamer', params={'opt': 'val'})
        - using 'problem_kind' parameter.
          e.g. Grounder(problem_kind=problem.kind)
        """
        return self._get_solver('grounder', name, None, params, problem_kind)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.test.examples import get_example_problems
from unified_planning.solvers import SequentialPlanValidator
from unified_planning.environment import get_env

class TestProblem(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_all(self):
        pv = SequentialPlanValidator(env=get_env())
        for p in self.problems.values():
            if p.problem.kind.has_continuous_time():
                continue
            problem, plan = p.problem, p.plan
            self.assertTrue(pv.validate(problem, plan))

    def test_all_from_factory(self):
        with PlanValidator(name='sequential_plan_validator') as pv:
            self.assertEqual(pv.name, 'sequential_plan_validator')
            for p in self.problems.values():
                if p.problem.kind.has_continuous_time():
                    continue
                problem, plan = p.problem, p.plan
                self.assertTrue(pv.validate(problem, plan))

    def test_all_from_factory_with_problem_kind(self):
        for p in self.problems.values():
            problem, plan = p.problem, p.plan
            if problem.kind.has_continuous_time() or \
               problem.kind.has_actions_cost() or \
               problem.kind.has_final_value():
                continue
            env = unified_planning.environment.Environment()
            env.factory.solvers.pop('tamer', None)
            with env.factory.PlanValidator(problem_kind=problem.kind) as pv:
                self.assertEqual(pv.name, 'sequential_plan_validator')
                self.assertTrue(pv.validate(problem, plan))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
'''
This module defines the Fluent class.
A Fluent has a name, a type and a signature
that defines the types of its parameters.
'''

import unified_planning as up
from unified_planning.environment import get_env, Environment
from typing import List, OrderedDict, Optional, Union


class Fluent:
    '''Represents a fluent.'''
    def __init__(self, name: str, typename: 'up.model.types.Type' = None,
                 _signature: Optional[Union[OrderedDict[str, 'up.model.types.Type'], List['up.model.parameter.Parameter']]] = None,
                 env: Environment = None, **kwargs: 'up.model.types.Type'):
        self._env = get_env(env)
        self._name = name
        if typename is None:
            self._typename = self._env.type_manager.BoolType()
        else:
            self._typename = typename
        self._signature: List['up.model.parameter.Parameter'] = []
        if _signature is not None:
            assert len(kwargs) == 0
            if isinstance(_signature, OrderedDict):
                for param_name, param_type in _signature.items():
                    self._signature.append(up.model.parameter.Parameter(param_name, param_type))
            elif isinstance(_signature, List):
                self._signature = _signature[:]
            else:
                raise NotImplementedError
        else:
            for param_name, param_type in kwargs.items():
                self._signature.append(up.model.parameter.Parameter(param_name, param_type))

    def __repr__(self) -> str:
        sign = ''
        if self.arity > 0:
            sign_items = [f'{p.name}={str(p.type)}' for p in self.signature]
            sign = f'[{", ".join(sign_items)}]'
        return f'{str(self.type)} {self.name}{sign}'

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Fluent):
            return self._name == oth._name and self._typename == oth._typename and self._signature == oth._signature and self._env == oth._env
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._typename)
        for p in self._signature:
            res += hash(p)
        return res ^ hash(self._name)

    @property
    def name(self) -> str:
        '''Returns the fluent name.'''
        return self._name

    @property
    def type(self) -> 'up.model.types.Type':
        '''Returns the fluent type.'''
        return self._typename

    @property
    def signature(self) -> List['up.model.parameter.Parameter']:
        '''Returns the fluent signature.
        The signature is the List of Parameters.
        '''
        return self._signature

    @property
    def arity(self) -> int:
        '''Returns the fluent arity.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        minimum time as possible.'''
        return len(self._signature)

    def __call__(self, *args: 'up.model.expression.Expression') -> 'up.model.fnode.FNode':
        '''Returns a fluent expression with the given parameters.'''
        return self._env.expression_manager.FluentExp(self, args)



import os
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.exceptions import UPUsageError
from unified_planning.model.problem_kind import basic_classical_kind, classical_kind, full_numeric_kind, basic_temporal_kind, hierarchical_kind
from unified_planning.test import TestCase, skipIfNoPlanValidatorForProblemKind, skipIfNoOneshotPlannerForProblemKind, skipIfSolverNotAvailable
from unified_planning.test.examples import get_example_problems
from unified_planning.transformers import Grounder as TransformersGrounder


class TestGrounder(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_basic(self):
        problem = self.problems['basic'].problem

        gro = TransformersGrounder(problem)
        with self.assertRaises(UPUsageError) as e:
            gro.get_rewrite_back_map()
        self.assertIn('The get_rewrite_back_map method must be called after the function get_rewritten_problem!', str(e.exception))

        grounded_problem = gro.get_rewritten_problem()
        grounded_problem_2 = gro.get_rewritten_problem()
        self.assertEqual(grounded_problem, grounded_problem_2)
        grounded_problem.name = problem.name
        self.assertEqual(grounded_problem, problem)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot(self):
        problem = self.problems['robot'].problem

        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 2)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = gro.rewrite_back_plan(grounded_plan)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected(self):
        problem = self.problems['robot_locations_connected'].problem

        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 28)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)
        for i, a in enumerate(problem.actions):
            if i == 0:
                self.assertEqual(len(gro.get_transformed_actions(a)), 8)
            elif i == 1:
                self.assertEqual(len(gro.get_transformed_actions(a)), 20)
            else:
                self.assertTrue(False)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = gro.rewrite_back_plan(grounded_plan)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected_from_factory(self):
        problem = self.problems['robot_locations_connected'].problem

        with Grounder(name='up_grounder') as embedded_grounder:
            self.assertTrue(embedded_grounder.supports(problem.kind))
            grounded_problem, rewrite_plan_funct = embedded_grounder.ground(problem)
            self.assertEqual(len(grounded_problem.actions), 28)
            for a in grounded_problem.actions:
                self.assertEqual(len(a.parameters), 0)

            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_plan_funct(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected_from_factory_with_problem_kind(self):
        problem = self.problems['robot_locations_connected'].problem
        kind = problem.kind

        with Grounder(problem_kind=kind) as embedded_grounder:
            self.assertTrue(embedded_grounder.supports(kind))
            grounded_problem, rewrite_plan_funct = embedded_grounder.ground(problem)
            self.assertEqual(len(grounded_problem.actions), 28)
            for a in grounded_problem.actions:
                self.assertEqual(len(a.parameters), 0)

            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_plan_funct(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(hierarchical_kind)
    @skipIfNoPlanValidatorForProblemKind(hierarchical_kind)
    def test_hierarchical_blocks_world(self):
        problem = self.problems['hierarchical_blocks_world'].problem

        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 108)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)
        for i, a in enumerate(problem.actions):
            if i == 0:
                self.assertEqual(len(gro.get_transformed_actions(a)), 108)
            else:
                self.assertTrue(False)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = gro.rewrite_back_plan(grounded_plan)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_matchcellar(self):
        problem = self.problems['matchcellar'].problem

        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 6)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)
        for a in problem.actions:
            self.assertEqual(len(gro.get_transformed_actions(a)), 3)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = gro.rewrite_back_plan(grounded_plan)
            for _, ai, _ in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_matchcellar_grounder_from_factory(self):
        problem = self.problems['matchcellar'].problem

        gro = TransformersGrounder(problem)
        grounded_problem_test = gro.get_rewritten_problem()
        with Grounder(name='up_grounder') as grounder:
            self.assertTrue(grounder.supports(problem.kind))
            grounded_problem_try, rewrite_back_plan_function = grounder.ground(problem)
            self.assertEqual(grounded_problem_test, grounded_problem_try)
            with OneshotPlanner(problem_kind=grounded_problem_try.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem_try).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for _, ai, _ in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    def test_timed_connected_locations(self):
        problem = self.problems['timed_connected_locations'].problem

        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 20)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)
        for a in problem.actions:
            self.assertEqual(len(gro.get_transformed_actions(a)), 20)

    def test_ad_hoc_1(self):
        problem = Problem('ad_hoc')
        Location = UserType('Location')
        visited = Fluent('at', BoolType(), position=Location)
        l1 = Object('l1', Location)
        visit = InstantaneousAction('visit', l_to=Location)
        l_to = visit.parameter('l_to')
        visit.add_effect(visited(l_to), True)
        visit_l1 = InstantaneousAction('visit_l1')
        visit_l1.add_effect(visited(l1), True)
        problem.add_fluent(visited)
        problem.set_initial_value(visited(l1), True)
        problem.add_object(l1)
        problem.add_action(visit)
        problem.add_action(visit_l1)
        gro = TransformersGrounder(problem)
        grounded_problem = gro.get_rewritten_problem()
        self.assertEqual(len(grounded_problem.actions), 2)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)
        for a in problem.actions:
            self.assertEqual(len(gro.get_transformed_actions(a)), 1)

    def test_ad_hoc_2(self):
        problem = Problem('ad_hoc')
        gro = TransformersGrounder(problem)
        gro.get_rewritten_problem()
        with self.assertRaises(NotImplementedError):
            gro.rewrite_back_plan(problem)

    @skipIfSolverNotAvailable('pyperplan')
    def test_pyperplan_grounder(self):

        problem = self.problems['robot_no_negative_preconditions'].problem
        for action in problem.actions:
            self.assertTrue(len(action.parameters) > 0)
        with Grounder(name='pyperplan') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfSolverNotAvailable('pyperplan')
    def test_pyperplan_grounder_mockup_problem(self):
        problem = Problem('mockup')
        Location = UserType('Location')
        at = Fluent('at', BoolType(), position=Location)
        at_l2 = Fluent('at_l2')
        l1 = Object('l1', Location)
        l2 = Object('l2', Location)
        move_to = InstantaneousAction('move_to', l_to=Location)
        l_to = move_to.parameter('l_to')
        move_to.add_effect(at(l_to), True)
        move_to_l2 = InstantaneousAction('move_to_l2')
        move_to_l2.add_effect(at_l2, True)
        problem.add_fluent(at, default_initial_value=False)
        problem.add_fluent(at_l2, default_initial_value=False)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.add_action(move_to)
        problem.add_action(move_to_l2)
        problem.add_goal(at(l1))
        problem.add_goal(at(l2))
        problem.add_goal(at_l2)

        with Grounder(name='pyperplan') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))



#!/usr/bin/env python3

from setuptools import setup, find_packages # type: ignore
import unified_planning


long_description=\
"""============================================================
 Unified planning: A library that unifies planning frameworks
 ============================================================
    Insert long description here
"""

setup(name='unified_planning',
      version=unified_planning.__version__,
      description='Unified Planning Framework',
      author='AIPlan4EU Project',
      author_email='aiplan4eu@fbk.eu',
      url='https://www.aiplan4eu-project.eu',
      packages=find_packages(),
      include_package_data=True,
      python_requires='>=3.7',  # supported Python ranges
      install_requires=['pyparsing'],
      extras_require={
          'dev':['tarski[arithmetic]','pytest','pytest-cov','mypy'],
          'tarski': ['tarski[arithmetic]']
      },
      license='APACHE',
      keywords='planning logic STRIPS RDDL',
      classifiers=[
          'Development Status :: 3 - Alpha',

          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',

          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'License :: OSI Approved :: Apache Software License',

          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
      ],
)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the quantifiers remover class."""


from unified_planning.walkers import ExpressionQuantifiersRemover
from unified_planning.transformers.transformer import Transformer
from unified_planning.model import Problem, InstantaneousAction, DurativeAction, Action
from typing import List, Dict


class QuantifiersRemover(Transformer):
    '''Quantifiers remover class:
    this class requires a problem and offers the capability
    to transform a problem with quantifiers into a problem without.
    '''
    def __init__(self, problem: Problem, name: str = 'qurm'):
        Transformer.__init__(self, problem, name)
        #NOTE no simplification are made. But it's possible to add them in key points
        self._expression_quantifier_remover = ExpressionQuantifiersRemover(self._env)
        #Represents the map from the new action to the old action
        self._new_to_old: Dict[Action, Action] = {}
        #represents a mapping from the action of the original problem to action of the new one.
        self._old_to_new: Dict[Action, List[Action]] = {}

    def get_rewritten_problem(self) -> Problem:
        '''Creates a problem that is a copy of the original problem
        but every quantifier is compiled away by it's respective logic formula
        instantiated on object.
        For example the formula:
            Forall (s-sempahore) is_green(s)
        in a problem with 3 objects of type semaphores {s1, s2, s3} is compiled in:
            And(is_green(s1), is_green(s2), is_green(s3)).
        While the respective formula on the same problem:
            Exists (s-semaphore) is_green(s)
        becomes:
            Or(is_green(s1), is_green(s2), is_green(s3)).'''
        if self._new_problem is not None:
            return self._new_problem
        #NOTE that a different environment might be needed when multi-threading
        self._new_problem = self._problem.clone()
        self._new_problem.name = f'{self._name}_{self._problem.name}'
        self._new_problem.clear_timed_goals()
        self._new_problem.clear_goals()
        for action in self._new_problem.actions:
            if isinstance(action, InstantaneousAction):
                original_action = self._problem.action(action.name)
                assert isinstance(original_action, InstantaneousAction)
                action.name = self.get_fresh_name(action.name)
                action.clear_preconditions()
                for p in original_action.preconditions:
                    action.add_precondition(self._expression_quantifier_remover.remove_quantifiers(p, self._problem))
                for e in action.effects:
                    if e.is_conditional():
                        e.set_condition(self._expression_quantifier_remover.remove_quantifiers(e.condition, self._problem))
                    e.set_value(self._expression_quantifier_remover.remove_quantifiers(e.value, self._problem))
                self._old_to_new[original_action] = [action]
                self._new_to_old[action] = original_action
            elif isinstance(action, DurativeAction):
                original_action = self._problem.action(action.name)
                assert isinstance(original_action, DurativeAction)
                action.name = self.get_fresh_name(action.name)
                action.clear_conditions()
                for i, cl in original_action.conditions.items():
                    for c in cl:
                        action.add_condition(i, self._expression_quantifier_remover.remove_quantifiers(c, self._problem))
                for t, el in action.effects.items():
                    for e in el:
                        if e.is_conditional():
                            e.set_condition(self._expression_quantifier_remover.remove_quantifiers(e.condition, self._problem))
                        e.set_value(self._expression_quantifier_remover.remove_quantifiers(e.value, self._problem))
                self._old_to_new[original_action] = [action]
                self._new_to_old[action] = original_action
            else:
                raise NotImplementedError
        for t, el in self._new_problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    e.set_condition(self._expression_quantifier_remover.remove_quantifiers(e.condition, self._problem))
                e.set_value(self._expression_quantifier_remover.remove_quantifiers(e.value, self._problem))
        for i, gl in self._problem.timed_goals.items():
            for g in gl:
                ng = self._expression_quantifier_remover.remove_quantifiers(g, self._problem)
                self._new_problem.add_timed_goal(i, ng)
        for g in self._problem.goals:
            ng = self._expression_quantifier_remover.remove_quantifiers(g, self._problem)
            self._new_problem.add_goal(ng)
        return self._new_problem

    def get_original_action(self, action: Action) -> Action:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the transformed problem into the actions of the original problem.'''
        return self._new_to_old[action]

    def get_transformed_actions(self, action: Action) -> List[Action]:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the original problem into the actions of the transformed problem.'''
        return self._old_to_new[action]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the dnf remover class."""

import unified_planning
from itertools import product
from unified_planning.model import FNode, Problem, InstantaneousAction, DurativeAction, TimeInterval, Timing, Action
from unified_planning.walkers import Dnf
from unified_planning.transformers.transformer import Transformer
from typing import List, Tuple, Dict


class DisjunctiveConditionsRemover(Transformer):
    '''DisjunctiveConditions remover class:
    this class requires a problem and offers the capability
    to transform a problem with preconditions not in the DNF form
    into one with all the preconditions in DNF form (an OR of AND);
    Then this OR is decomposed into subactions, therefore after this
    remover is called, every action condition or precondition will be
    an AND of leaf nodes.
    '''
    def __init__(self, problem: Problem, name: str = 'djrm'):
        Transformer.__init__(self, problem, name)
        #Represents the map from the new action to the old action
        self._new_to_old: Dict[Action, Action] = {}
        #represents a mapping from the action of the original problem to action of the new one.
        self._old_to_new: Dict[Action, List[Action]] = {}

    def get_rewritten_problem(self) -> Problem:
        '''Creates a problem that is a copy of the original problem
        but every precondition is compiled into one with preconditions in DNF.'''
        if self._new_problem is not None:
            return self._new_problem
        #NOTE that a different environment might be needed when multi-threading
        self._new_problem = self._problem.clone()
        self._new_problem.name = f'{self._name}_{self._problem.name}'
        self._new_problem.clear_actions()
        self._handle_actions()

        return self._new_problem

    def _handle_actions(self):
        dnf = Dnf(self._env)
        for a in self._problem.actions:
            if isinstance(a, InstantaneousAction):
                new_precond = dnf.get_dnf_expression(self._env.expression_manager.And(a.preconditions))
                if new_precond.is_or():
                    for and_exp in new_precond.args:
                        na = self._create_new_action_with_given_precond(and_exp, a)
                        self._new_problem.add_action(na)
                else:
                    na = self._create_new_action_with_given_precond(new_precond, a)
                    self._new_problem.add_action(na)
            elif isinstance(a, DurativeAction):
                interval_list: List[TimeInterval] = []
                conditions: List[List[FNode]] = []
                # save the timing, calculate the dnf of the and of all the conditions at the same time
                # and then save it in conditions.
                # conditions contains lists of Fnodes, where [a,b,c] means a or b or c
                for i, cl in a.conditions.items():
                    interval_list.append(i)
                    new_cond = dnf.get_dnf_expression(self._env.expression_manager.And(cl))
                    if new_cond.is_or():
                        conditions.append(new_cond.args)
                    else:
                        conditions.append([new_cond])
                conditions_tuple: Tuple[List[FNode], ...] = product(*conditions)
                for cond_list in conditions_tuple:
                    nda = self._create_new_durative_action_with_given_conds_at_given_times(interval_list, cond_list, a)
                    self._new_problem.add_action(nda)
            else:
                raise NotImplementedError

    def _create_new_durative_action_with_given_conds_at_given_times(self, interval_list: List[TimeInterval], cond_list: List[FNode], original_action: DurativeAction) -> DurativeAction:
        new_action = original_action.clone()
        new_action.name = self.get_fresh_name(original_action.name)
        new_action.clear_conditions()
        for i, c in zip(interval_list, cond_list):
            if c.is_and():
                for co in c.args:
                    new_action.add_condition(i, co)
            else:
                new_action.add_condition(i, c)
        assert self._new_to_old is not None
        self._new_to_old[new_action] = original_action
        self._map_old_to_new_action(original_action, new_action)
        return new_action

    def _create_new_action_with_given_precond(self, precond: FNode, original_action: InstantaneousAction) -> InstantaneousAction:
        new_action = original_action.clone()
        new_action.name = self.get_fresh_name(original_action.name)
        new_action.clear_preconditions()
        if precond.is_and():
            for leaf in precond.args:
                new_action.add_precondition(leaf)
        else:
            new_action.add_precondition(precond)
        assert self._new_to_old is not None
        self._new_to_old[new_action] = original_action
        self._map_old_to_new_action(original_action, new_action)
        return new_action

    def _map_old_to_new_action(self, old_action, new_action):
        if old_action in self._old_to_new:
            self._old_to_new[old_action].append(new_action)
        else:
            self._old_to_new[old_action] = [new_action]

    def get_original_action(self, action: Action) -> Action:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the transformed problem into the actions of the original problem.'''
        return self._new_to_old[action]

    def get_transformed_actions(self, action: Action) -> List[Action]:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the original problem into the actions of the transformed problem.'''
        return self._old_to_new[action]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from typing import Dict, Union, List, Set, Optional
from itertools import product

import unified_planning.environment
import unified_planning.solvers as solvers
import unified_planning.walkers as walkers
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model import FNode, Expression, Problem, ProblemKind, Object
from unified_planning.plan import SequentialPlan

class QuantifierSimplifier(walkers.Simplifier):
    """Same to the unified_planning.Simplifier, but does not expand quantifiers and solves them locally."""
    def __init__(self, env: 'unified_planning.environment.Environment', problem: Problem):
        walkers.DagWalker.__init__(self, True)
        self._env = env
        self.manager = env.expression_manager
        self._problem = problem
        self._assignments: Optional[Dict[Expression, Expression]] = None
        self._variable_assignments: Optional[Dict[Expression, Expression]] = None

    def qsimplify(self, expression: FNode, assignments: Dict[Expression, Expression], variable_assignments: Dict[Expression, Expression]):
        assert self._problem is not None
        assert self._assignments is None
        assert self._variable_assignments is None
        self._assignments = assignments
        self._variable_assignments = variable_assignments
        r = self.walk(expression)
        self._assignments = None
        self._variable_assignments = None
        return r

    def _push_with_children_to_stack(self, expression: FNode, **kwargs):
        """Add children to the stack."""
        if expression.is_forall() or expression.is_exists():
            self.stack.append((True, expression))
        else:
            super(walkers.Simplifier, self)._push_with_children_to_stack(expression, **kwargs)

    def _compute_node_result(self, expression: FNode, **kwargs):
        """Apply function to the node and memoize the result.
        Note: This function assumes that the results for the children
              are already available.
        """
        key = self._get_key(expression, **kwargs)
        if key not in self.memoization:
            try:
                f = self.functions[expression.node_type]
            except KeyError:
                f = self.walk_error

            if not (expression.is_forall() or expression.is_exists()):
                args = [self.memoization[self._get_key(s, **kwargs)] \
                        for s in self._get_children(expression)]
                self.memoization[key] = f(expression, args=args, **kwargs)
            else:
                self.memoization[key] = f(expression, args=expression.args, **kwargs)
        else:
            pass

    def _deep_subs_simplify(self, expression: FNode, variables_assignments: Dict[Expression, Expression]) -> FNode:
        new_qsimplifier = QuantifierSimplifier(self._env, self._problem)
        assert self._variable_assignments is not None
        assert self._assignments is not None
        copy = self._variable_assignments.copy()
        copy.update(variables_assignments)
        r = new_qsimplifier.qsimplify(expression, self._assignments, copy)
        assert r.is_constant()
        return r

    def walk_exists(self, expression: FNode, args: List[FNode]) -> FNode:
        assert self._problem is not None
        assert len(args) == 1
        if args[0].is_bool_constant():
            if args[0].bool_constant_value():
                return self.manager.TRUE()
            return self.manager.FALSE()
        vars = expression.variables()
        type_list = [v.type for v in vars]
        possible_objects: List[List[Object]] = [list(self._problem.objects_hierarchy(t)) for t in type_list]
        #product of n iterables returns a generator of tuples where
        # every tuple has n elements and the tuples make every possible
        # combination of 1 item for each iterable. For example:
        #product([1,2], [3,4], [5,6], [7]) =
        # (1,3,5,7) (1,3,6,7) (1,4,5,7) (1,4,6,7) (2,3,5,7) (2,3,6,7) (2,4,5,7) (2,4,6,7)
        for o in product(*possible_objects):
            subs: Dict[Expression, Expression] = dict(zip(vars, list(o)))
            result = self._deep_subs_simplify(args[0], subs)
            assert result.is_bool_constant()
            if result.bool_constant_value():
                return self.manager.TRUE()
        return self.manager.FALSE()

    def walk_forall(self, expression: FNode, args: List[FNode]) -> FNode:
        assert self._problem is not None
        assert len(args) == 1
        if args[0].is_bool_constant():
            if args[0].bool_constant_value():
                return self.manager.TRUE()
            return self.manager.FALSE()
        vars = expression.variables()
        type_list = [v.type for v in vars]
        possible_objects: List[List[Object]] = [list(self._problem.objects_hierarchy(t)) for t in type_list]
        #product of n iterables returns a generator of tuples where
        # every tuple has n elements and the tuples make every possible
        # combination of 1 item for each iterable. For example:
        #product([1,2], [3,4], [5,6], [7]) =
        # (1,3,5,7) (1,3,6,7) (1,4,5,7) (1,4,6,7) (2,3,5,7) (2,3,6,7) (2,4,5,7) (2,4,6,7)
        for o in product(*possible_objects):
            subs: Dict[Expression, Expression] = dict(zip(vars, list(o)))
            result = self._deep_subs_simplify(args[0], subs)
            assert result.is_bool_constant()
            if not result.bool_constant_value():
                return self.manager.FALSE()
        return self.manager.TRUE()

    def walk_fluent_exp(self, expression: FNode, args: List[FNode]) -> FNode:
        new_exp = self.manager.FluentExp(expression.fluent(), tuple(args))
        assert self._assignments is not None
        res = self._assignments.get(new_exp, None)
        if res is not None:
            res, = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(f"Value of Fluent {str(expression)} not found in {str(self._assignments)}")

    def walk_variable_exp(self, expression: FNode, args: List[FNode]) -> FNode:
        assert self._variable_assignments is not None
        res = self._variable_assignments.get(expression.variable(), None)
        if res is not None:
            res, = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(f"Value of Variable {str(expression)} not found in {str(self._variable_assignments)}")

    def walk_param_exp(self, expression: FNode, args: List[FNode]) -> FNode:
        assert self._assignments is not None
        res = self._assignments.get(expression.parameter(), None)
        if res is not None:
            res, = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(f"Value of Parameter {str(expression)} not found in {str(self._assignments)}")


class SequentialPlanValidator(solvers.solver.Solver):
    """Performs plan validation."""
    def __init__(self, **options):
        self._env: 'unified_planning.environment.Environment' = unified_planning.environment.get_env(options.get('env', None))
        self.manager = self._env.expression_manager
        self._substituter = walkers.Substituter(self._env)
        self._last_error: Union[str, None] = None

    def validate(self, problem: 'Problem', plan: 'unified_planning.plan.Plan') -> bool:
        """Returns True if and only if the plan given in input is a valid plan for the problem given in input.
        This means that from the initial state of the problem, by following the plan, you can reach the
        problem goal. Otherwise False is returned."""
        assert isinstance(plan, SequentialPlan)
        self._qsimplifier = QuantifierSimplifier(self._env, problem)
        self._last_error = None
        assignments: Dict[Expression, Expression] = problem.initial_values.copy() # type: ignore
        count = 0 #used for better error indexing
        for ai in plan.actions:
            action = ai.action
            assert isinstance(action, unified_planning.model.InstantaneousAction)
            count = count + 1
            new_assignments: Dict[Expression, Expression] = {}
            for ap, oe in zip(ai.action.parameters, ai.actual_parameters):
                assignments[ap] = oe
            for p in action.preconditions:
                ps = self._subs_simplify(p, assignments)
                if not (ps.is_bool_constant() and ps.bool_constant_value()):
                    self._last_error = f'Precondition {p} of {str(count)}-th action instance {str(ai)} is not satisfied.'
                    return False
            for e in action.effects:
                cond = True
                if e.is_conditional():
                    ec = self._subs_simplify(e.condition, assignments)
                    assert ec.is_bool_constant()
                    cond = ec.bool_constant_value()
                if cond:
                    ge = self._get_ground_fluent(e.fluent, assignments)
                    if e.is_assignment():
                        new_assignments[ge] = self._subs_simplify(e.value, assignments)
                    elif e.is_increase():
                        new_assignments[ge] = self._subs_simplify(self.manager.Plus(e.fluent,
                                                e.value), assignments)
                    elif e.is_decrease():
                        new_assignments[ge] = self._subs_simplify(self.manager.Minus(e.fluent,
                                                e.value), assignments)
            assignments.update(new_assignments)
            for ap in action.parameters:
                del assignments[ap]
        for g in problem.goals:
            gs = self._subs_simplify(g, assignments)
            if not (gs.is_bool_constant() and gs.bool_constant_value()):
                    self._last_error = f'Goal {str(g)} is not reached by the plan.'
                    return False
        return True

    def get_last_error_info(self) -> str:
        """When the function 'is_valid_plan()' is called, if the result is false
        this function returns a string containing information on why the plan is not valid."""
        assert not self._last_error is None
        return self._last_error

    def _get_ground_fluent(self, fluent:FNode, assignments: Dict[Expression, Expression]) -> FNode:
        assert fluent.is_fluent_exp()
        new_args = []
        for p in fluent.args:
            new_args.append(self._subs_simplify(p, assignments))
        return self.manager.FluentExp(fluent.fluent(), tuple(new_args))

    def _subs_simplify(self, expression: FNode, assignments: Dict[Expression, Expression]) -> FNode:
        r = self._qsimplifier.qsimplify(expression, assignments, {})
        assert r.is_constant()
        return r

    @property
    def name(self):
        return 'sequential_plan_validator'

    @staticmethod
    def supports(problem_kind):
        supported_kind = ProblemKind()
        supported_kind.set_typing('FLAT_TYPING')
        supported_kind.set_typing('HIERARCHICAL_TYPING')
        supported_kind.set_numbers('CONTINUOUS_NUMBERS')
        supported_kind.set_numbers('DISCRETE_NUMBERS')
        supported_kind.set_conditions_kind('NEGATIVE_CONDITIONS')
        supported_kind.set_conditions_kind('DISJUNCTIVE_CONDITIONS')
        supported_kind.set_conditions_kind('EQUALITY')
        supported_kind.set_conditions_kind('EXISTENTIAL_CONDITIONS')
        supported_kind.set_conditions_kind('UNIVERSAL_CONDITIONS')
        supported_kind.set_effects_kind('CONDITIONAL_EFFECTS')
        supported_kind.set_effects_kind('INCREASE_EFFECTS')
        supported_kind.set_effects_kind('DECREASE_EFFECTS')
        supported_kind.set_fluents_type('NUMERIC_FLUENTS')
        supported_kind.set_fluents_type('OBJECT_FLUENTS')
        return problem_kind <= supported_kind

    @staticmethod
    def is_plan_validator():
        return True

    def destroy(self):
        pass



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Effect class.
A basic Effect has a fluent and an expression.
A condition can be added to make it a conditional effect.
"""


from unified_planning.model.fnode import FNode
from enum import Enum, auto


class EffectKind(Enum):
    ASSIGN = auto()
    INCREASE = auto()
    DECREASE = auto()

class Effect:
    def __init__(self, fluent: FNode, value: FNode, condition: FNode, kind: EffectKind = EffectKind.ASSIGN):
        self._fluent = fluent
        self._value = value
        self._condition = condition
        self._kind = kind

    def __repr__(self) -> str:
        s = []
        if self.is_conditional():
            s.append(f'if {str(self._condition)} then')
        s.append(f'{str(self._fluent)}')
        if self.is_assignment():
            s.append(':=')
        elif self.is_increase():
            s.append('+=')
        elif self.is_decrease():
            s.append('-=')
        s.append(f'{str(self._value)}')
        return ' '.join(s)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Effect):
            return self._fluent == oth._fluent and self._value == oth._value and self._condition == oth._condition and self._kind == oth._kind
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._fluent) + hash(self._value) + hash(self._condition) + hash(self._kind)

    def clone(self):
        new_effect = Effect(self._fluent, self._value, self._condition, self._kind)
        return new_effect

    def is_conditional(self) -> bool:
        """Returns True if the Effect condition is not True."""
        return not self._condition.is_true()

    @property
    def fluent(self) -> FNode:
        """Returns the Fluent that is modified by this effect."""
        return self._fluent

    @property
    def value(self) -> FNode:
        """Returns the value given to the Fluent by this Effect."""
        return self._value

    def set_value(self, new_value: FNode):
        """Sets the value given to the Fluent by this Effect."""
        self._value = new_value

    @property
    def condition(self) -> FNode:
        """Returns the condition required for this Effect to be applied."""
        return self._condition

    def set_condition(self, new_condition: FNode):
        """Sets the condition required for this Effect to be applied."""
        self._condition = new_condition

    @property
    def kind(self) -> EffectKind:
        """Returns the kind of this Effect."""
        return self._kind

    def is_assignment(self) -> bool:
        """Returns True if the kind of this Effect is an assignment."""
        return self._kind == EffectKind.ASSIGN

    def is_increase(self) -> bool:
        """Returns True if the kind of this Effect is an increase."""
        return self._kind == EffectKind.INCREASE

    def is_decrease(self) -> bool:
        """Returns True if the kind of this Effect is a decrease."""
        return self._kind == EffectKind.DECREASE



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
from typing import List, Optional, Union

import unified_planning
import unified_planning.model
import unified_planning.walkers as walkers

import tarski # type: ignore

class TarskiFormulaConverter(walkers.DagWalker):
    def __init__(self, language: 'tarski.fol.FirstOrderLanguage', env, object_freshname: str) -> None:
        walkers.DagWalker.__init__(self)
        self.object_freshname = object_freshname
        self.lang = language
        self.env = env

    def convert_formula(self, expression: 'unified_planning.model.FNode') -> 'tarski.syntax.formulas.Formula':
        return self.walk(expression)

    def walk_and(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        return tarski.syntax.land(*args, flat=True)

    def walk_or(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        return tarski.syntax.lor(*args, flat=True)

    def walk_not(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 1
        return tarski.syntax.neg(args[0])

    def walk_iff(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return tarski.syntax.equiv(args[0], args[1])

    def walk_implies(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return tarski.syntax.implies(args[0], args[1])

    def walk_exists(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 1
        variables = [self.lang.variable(v.name, self.lang.get_sort(v.type.name if v.type.name != 'object' else self.object_freshname)) for v in expression.variables()] # type: ignore
        return tarski.syntax.exists(*variables, args[0])

    def walk_forall(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 1
        variables = [self.lang.variable(v.name, self.lang.get_sort(v.type.name if v.type.name != 'object' else self.object_freshname)) for v in expression.variables()] # type: ignore
        return tarski.syntax.forall(*variables, args[0])

    def walk_equals(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return tarski.syntax.Atom(self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.EQ), args)

    def walk_le(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return tarski.syntax.Atom(self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.LE), args)

    def walk_lt(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return tarski.syntax.Atom(self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.LT), args)

    def walk_fluent_exp(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        tarski_fluent_rep = self.lang.get(expression.fluent().name)
        new_args = []
        for i, x in enumerate(expression.args):
            if x.is_int_constant():
                type = expression.fluent().signature[i].type
                typename = _type_name_added_to_language_if_needed(self.lang, type)
                constant = tarski.syntax.Constant(x.int_constant_value(), \
                    self.lang.get_sort(typename))
                new_args.append(constant)
            elif x.is_real_constant():
                raise unified_planning.exceptions.UPProblemDefinitionError('Fluents can not have reals into their signatures.')
            else:
                new_args.append(args[i])
        return tarski_fluent_rep(*new_args)

    def walk_plus(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        value = args[0]
        for a in args[1:]:
            value = value + a
        return value

    def walk_minus(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return args[0] - args[1]

    def walk_times(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        value = args[0]
        for a in args[1:]:
            value = value * a
        return value

    def walk_div(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 2
        return args[0] / args[1]

    def walk_bool_constant(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 0
        if expression.bool_constant_value():
            return tarski.syntax.top
        return tarski.syntax.bot

    def walk_int_constant(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 0
        return tarski.syntax.Constant(expression.int_constant_value(), self.lang.Integer)

    def walk_real_constant(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        assert len(args) == 0
        return tarski.syntax.Constant(expression.real_constant_value(), self.lang.Real)

    def walk_param_exp(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        type = expression.parameter().type
        if type.is_user_type():
            return self.lang.variable(expression.parameter().name, \
                self.lang.get_sort(type.name if type.name != 'object' else self.object_freshname)) # type: ignore
        else:
            return self.lang.variable(expression.parameter().name, \
                self.lang.get_sort(_type_name_added_to_language_if_needed(self.lang, type)))

    def walk_variable_exp(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        type = expression.variable().type
        if type.is_user_type():
            return self.lang.variable(expression.variable().name, \
                self.lang.get_sort(type.name if type.name != 'object' else self.object_freshname)) # type: ignore
        else:
            return self.lang.variable(expression.variable().name, \
                self.lang.get_sort(_type_name_added_to_language_if_needed(self.lang, type)))

    def walk_object_exp(self, expression: 'unified_planning.model.FNode', args: List['tarski.syntax.formulas.Formula']) -> 'tarski.syntax.formulas.Formula':
        return self.lang.get_constant(expression.object().name)

def convert_problem_to_tarski(problem: 'unified_planning.model.Problem') -> 'tarski.fstrips.problem.Problem':
    '''Converts a problem in the unified_planning.model.Problem representation in the equivalent
    tarski.fstrips.Problem representation.'''
    features: List[str] = []
    kind = problem.kind
    if kind.has_equality(): # type: ignore
        features.append('equality')
    if kind.has_continuous_numbers() or kind.has_discrete_numbers() or kind.has_numeric_fluents(): # type: ignore
        features.append('arithmetic')
    object_freshname = 'object_'
    while problem.has_type(object_freshname):
        object_freshname = object_freshname + '_'
    #creating tarski language
    lang = tarski.fstrips.language(f'{problem.name}_lang', features)
    for ut in problem.user_types: #adding user_types to the language
        if ut.father is not None:  # type: ignore
            lang.sort(ut.name if ut.name != 'object' else object_freshname, ut.father.name if ut.father.name != 'object' else object_freshname) # type: ignore
        else:
            lang.sort(ut.name if ut.name != 'object' else object_freshname) # type: ignore
    for fluent in problem.fluents: #adding fluents to the language
        signature = []
        for param in fluent.signature:
            if param.type.is_user_type():
                signature.append(lang.get_sort(param.type.name if param.type.name != 'object' else object_freshname)) # type: ignore
            else:
                #typename will be the name that this type has in the tarski language
                typename = _type_name_added_to_language_if_needed(lang, param.type)
                signature.append(lang.get_sort(typename))
        if fluent.type.is_bool_type():
            lang.predicate(fluent.name, *signature)
        else:
            typename = _type_name_added_to_language_if_needed(lang, fluent.type)
            lang.function(fluent.name, *signature, lang.get_sort(typename if typename != 'object' else object_freshname))
    for o in problem.all_objects: #adding objects to the language
        lang.constant(o.name, lang.get_sort(o.type.name if o.type.name != 'object' else object_freshname)) # type: ignore
    #creating tarski problem
    em = problem.env.expression_manager
    tfc = TarskiFormulaConverter(language=lang, env=problem.env, object_freshname=object_freshname)
    new_problem = tarski.fstrips.problem.create_fstrips_problem(lang, problem.name, f'{problem.name}_domain')
    for action in problem.actions:
        if not isinstance(action, unified_planning.model.InstantaneousAction):
            raise unified_planning.exceptions.UPProblemDefinitionError('Tarski supports only Instantaneous Actions.')
        parameters = []
        for p in action.parameters:
            if p.type.is_user_type():
                parameters.append(lang.variable(p.name, lang.get_sort(p.type.name if p.type.name != 'object' else object_freshname))) # type: ignore
            else:
                parameters.append(lang.variable(p.name, lang.get_sort(_type_name_added_to_language_if_needed(lang, p.type))))
        #add action to the problem
        new_problem.action(action.name,
                            parameters,
                            precondition=tfc.convert_formula(em.And(action.preconditions)),
                            effects=[_convert_effect(e, tfc, em) for e in action.effects])
    for fluent_exp, value_exp in problem.initial_values.items():
        if value_exp.is_bool_constant():
            if value_exp.constant_value():
                parameters = []
                for a in fluent_exp.args:
                    parameters.append(tfc.convert_formula(a))
                new_problem.init.add(lang.get_predicate(fluent_exp.fluent().name), *parameters)
        else:
            value: Optional[Union[Fraction, int, 'tarski.syntax.formulas.Formula']] = None
            if value_exp.is_int_constant():
                value = value_exp.int_constant_value()
            elif value_exp.is_real_constant():
                value = value_exp.real_constant_value()
            else:
                value = tfc.convert_formula(value_exp)
            new_problem.init.set(tfc.convert_formula(fluent_exp), value)
    new_problem.goal = tfc.convert_formula(em.And(problem.goals))
    return new_problem

def _type_name_added_to_language_if_needed(lang: 'tarski.FirstOrderLanguage', type: 'unified_planning.model.Type') -> str:
    typename = ''
    if type.is_user_type():
        typename = type.name # type: ignore
    else:
        typename = str(type).replace(' ','')
    if type.is_int_type() and (type.lower_bound is None or type.upper_bound is None): # type: ignore
        if type.lower_bound is None and type.upper_bound is None: # type: ignore
            typename = 'Integer'
        elif type.lower_bound == 0: # type: ignore
            assert type.upper_bound is None # type: ignore #must be true, otherwise branch would be skipped
            typename = 'Natural'
        else:
            raise unified_planning.exceptions.UPProblemDefinitionError('Int type with just one bound is not accepted by tarski')
    elif type.is_real_type() and (type.lower_bound is None or type.upper_bound is None): # type: ignore
        if type.lower_bound is None and type.upper_bound is None: # type: ignore
            typename = 'Real'
        else:
            raise unified_planning.exceptions.UPProblemDefinitionError('Real type with just one bound is not accepted by tarski')
    if not lang.has_sort(typename):
        # the type is not in the language, therefore it must be added
        if type.is_int_type():
            lang.interval(typename, lang.Integer, type.lower_bound, type.upper_bound) # type: ignore
        elif type.is_real_type():
            lang.interval(typename, lang.Real, type.lower_bound, type.upper_bound) # type: ignore
        else:
            raise NotImplementedError
    return typename

def _convert_effect(effect: 'unified_planning.model.Effect', tfc: TarskiFormulaConverter, em: 'unified_planning.model.ExpressionManager') -> 'tarski.fstrips.SingleEffect':
    condition = tfc.convert_formula(effect.condition)
    predicate = tfc.convert_formula(effect.fluent)
    if effect.value.is_bool_constant():
        if effect.value.constant_value():
            return tarski.fstrips.AddEffect(predicate, condition)
        else:
            return tarski.fstrips.DelEffect(predicate, condition)
    else:
        if effect.kind == unified_planning.model.EffectKind.ASSIGN:
            value = tfc.convert_formula(effect.value)
        elif effect.kind == unified_planning.model.EffectKind.INCREASE:
            value = tfc.convert_formula(em.Plus(effect.fluent, effect.value))
        elif effect.kind == unified_planning.model.EffectKind.DECREASE:
            value = tfc.convert_formula(em.Minus(effect.fluent, effect.value))
        else:
            raise NotImplementedError
        return tarski.fstrips.FunctionalEffect(predicate, value, condition)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from typing import Dict, Optional


class PlanQualityMetric():
    '''This is the base class of any metric for plan quality'''
    pass

class MinimizeActionCosts(PlanQualityMetric):
    def __init__(self, costs: Dict['up.model.Action', 'up.model.FNode'],
                 default: Optional['up.model.FNode'] = None):
        self.costs = costs
        self.default = default

    def __repr__(self):
        costs = {a.name: c for a, c in self.costs.items()}
        costs['default'] = self.default
        return f'minimize actions-cost: {costs}'

    def get_action_cost(self, action: 'up.model.Action') -> Optional['up.model.FNode']:
        return self.costs.get(action, self.default)

class MinimizeSequentialPlanLength(PlanQualityMetric):
    def __repr__(self):
        return 'minimize sequential-plan-length'

class MinimizeMakespan(PlanQualityMetric):
    def __repr__(self):
        return 'minimize makespan'

class MinimizeExpressionOnFinalState(PlanQualityMetric):
    def __init__(self, expression: 'up.model.FNode'):
        self.expression = expression

    def __repr__(self):
        return f'minimize {self.expression}'

class MaximizeExpressionOnFinalState(PlanQualityMetric):
    def __init__(self, expression: 'up.model.FNode'):
        self.expression = expression

    def __repr__(self):
        return f'maximize {self.expression}'



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase
from unified_planning.test.examples import get_example_problems
from unified_planning.transformers.transformer import Transformer

class TestQuantifiersRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_transformer(self):
        problem = self.problems['basic'].problem
        a = problem.action('a')
        t = Transformer(problem, 't')
        t._new_problem = problem
        self.assertEqual(t.get_fresh_name(a.name), 't_a_0')
        with self.assertRaises(NotImplementedError):
            t.get_original_action(a)
        with self.assertRaises(NotImplementedError):
            t.get_transformed_actions(a)
        with self.assertRaises(NotImplementedError):
            t.get_rewritten_problem()
        with self.assertRaises(NotImplementedError):
            t.rewrite_back_plan(None) # type: ignore



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the solver interface."""


import unified_planning as up
from unified_planning.plan import Plan
from unified_planning.model import ProblemKind, Problem
from enum import Enum, auto
from typing import IO, Optional, Tuple, Callable, Union



class OptimalityGuarantee(Enum):
    SATISFICING = auto()
    SOLVED_OPTIMALLY = auto()


class Solver:
    """Represents the solver interface."""

    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise

    @property
    def name(self) -> str:
        raise NotImplementedError

    @staticmethod
    def is_oneshot_planner() -> bool:
        return False

    @staticmethod
    def satisfies(optimality_guarantee: Union[OptimalityGuarantee, str]) -> bool:
        return False

    @staticmethod
    def is_plan_validator() -> bool:
        return False

    @staticmethod
    def is_grounder() -> bool:
        return False

    @staticmethod
    def supports(problem_kind: 'ProblemKind') -> bool:
        return len(problem_kind.features) == 0

    def solve(self, problem: 'up.model.Problem',
                    callback: Optional[Callable[['up.solvers.results.PlanGenerationResult'], None]] = None,
                    timeout: Optional[float] = None,
                    output_stream: Optional[IO[str]] = None) -> 'up.solvers.results.PlanGenerationResult':
        '''This method takes a up.model.Problem and returns a up.solvers.results.PlanGenerationResult,
        which contains information about the solution to the problem given by the planner.

        :param problem: is the up.model.Problem to solve.
        :param callback: is a function used by the planner to give reports to the user during the problem resolution, defaults to None.
        :param timeout: is the time in seconds that the planner has at max to solve the problem, defaults to None.
        :param output_stream: is a stream of strings where the planner writes his
        output (and also errors) while the planner is solving the problem, defaults to None
        :return: the up.solvers.results.PlanGenerationResult created by the planner; a data structure containing the up.plan.Plan found
        and some additional information about it.

        The only required parameter is "problem" but the planner should warn the user if callback, timeout or
        output_stream are not None and the planner ignores them.'''
        raise NotImplementedError

    def validate(self, problem: 'up.model.Problem', plan: 'up.plan.Plan') -> bool:
        raise NotImplementedError

    def ground(self, problem: 'up.model.Problem') -> Tuple[Problem, Callable[[Plan], Plan]]:
        '''
        Implement only if "self.is_grounder()" returns True.
        This function should return the tuple (grounded_problem, trace_back_plan), where
        "trace_back_plan" is a callable from a plan for the "grounded_problem" to a plan of the
        original problem.

        NOTE: to create a callable, the "functools.partial" method can be used, as we do in the
        "up.solvers.grounder".

        Also, the "up.solvers.grounder.lift_plan" function can be called, if retrieving the needed map
        fits the solver implementation better than retrieving a function.'''
        raise NotImplementedError

    def destroy(self):
        raise NotImplementedError

    def __enter__(self):
        """Manages entering a Context (i.e., with statement)"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Manages exiting from Context (i.e., with statement)"""
        self.destroy()
